#!/usr/bin/env python
# coding=utf-8

## A standard launcher plugin. 
## The file name for a launcher-plugin has always to begin with "launch", folowed by any fancy name.
## Please have a look at the Plugins-list at createweb.de/en/webmarks to avoid name-collisions 
## with plugins that already exist!!! 
##
## There has always to be a main class called "Plugin" (case-sensitive).
## This class must include the methods "launch" and "get_name" as described above.
## The Launch command must be run as thread, or independent process,
## otherwise the entire program will freeze until the launched application is closed!
######
## If you write new Plugins, please leave all initial comments untouched and comment
## your own code as well, so that other developers may easily derive new plugins from it.
##Please register your plugin at createweb.de/en/webmarks, even
## If you'll publish it on your own site or elsewhere, to keep the naming of all 
## existing plugins consistent.
##
##	This plugin is free software; you can redistribute it and/or
##	modify it under the terms of the GNU Lesser General Public License
##	as published by the Free Software Foundation; either version 2
##	of the License, or (at your option) any later version.
##	
##	This program is distributed in the hope that it will be useful,
##	but WITHOUT ANY WARRANTY; without even the implied warranty of
##	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##	GNU General Public License for more details.
##	
##	You should have received a copy of the GNU Lesser General Public License
##	along with this program; if not, write to the Free Software
##	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

'''Utilities for opening files or URLs in the registered default application
and for sending e-mail using the user's preferred composer.

'''

__version__ = '1.0'
__all__ = ['open', 'mailto']

import os
import sys
import webbrowser
import subprocess
import gettext
import webmarx

## get the gettext translation from webmarx.
## NEVER REMOVE THIS, ELSE TRANSLATIONS WILL BE BROKEN!
_ = webmarx._

from email.Utils import encode_rfc2231

_controllers = {}
_open = None
	
class Plugin:
	''' A standard launcher plugin.
	The file-name for a launcher-plugin has allways to begin with
	
	'''
	def __init__(self):
		''' Initialise the plugin.
		
		name is the string that will be displayed on the launch-button.
		it should always be in brackets with a leading underline to be
		translatable. You may call this attribute other than name 
		or even generate it from a function, as long as
		you provide the proper string through the "get_name" method.
		
		'''		
		self.name = _('Default application')
		self.description = _('Open the bookmark with your systems default application \
		fitting for the type and protocol of the url.')
		self.enabled = True
		self.haveconfig = False
				
		self.show_button = True # Show the launch button for this plugin by default
		self.show_edit_combo = True # Show the entry in the editors combobox-default-app for this plugin by default

		
	def launch(self, url):
		''' This method is mandatory for all launcher plugins. 
		it has always to receive the url as 2. argument.
		
		Return	boolean True | False (mandatory!)
		
		'''
		#print 'url: %s' % url
		if url == '':
			return False
		runit = openfile(url)
		if runit != True:
			print 'Error!'
		return True
	
	def get_name(self):
		''' This method is mandatory for all launcher plugins. 
		it has always to return the Name to be shown on the button as 
		translatable string (see above in __init__). The name has also to be unique 
		within to all installed plugins!!!
		'''
		return self.name
	
	def get_tooltip(self):
		''' This method is mandatory for all launcher plugins. 
		it has always to return a tooltip text for the button as 
		translatable string (see above in __init__), or None
		'''
		return self.description
		
	def open_config(self):
		''' NOT IMPLEMENTED YET FOR LAUNCHER PLUGINS!!!
		This method may be used to open a configuration-dialog for your plugin. 
		If a configuration is needed for your plugin, you have to write your own
		gtk-gui that will be opened by this method. You also have to set up your 
		own configuration-mechanism. There is no standardized way for this.
		If you want to use a configuration, set the parameter haveconfig in __init__
		to True, to show the config-button in the settings-dialog.
		
		'''
		return None


class BaseController(object):
	'''Base class for open program controllers.'''

	def __init__(self, name):
		self.name = name

	def open(self, filename):
		raise NotImplementedError


class Controller(BaseController):
	'''Controller for a generic open program.'''

	def __init__(self, *args):
		super(Controller, self).__init__(os.path.basename(args[0]))
		self.args = list(args)

	def _invoke(self, cmdline):
		if sys.platform[:3] == 'win':
			closefds = False
			startupinfo = subprocess.STARTUPINFO()
			startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
		else:
			closefds = True
			startupinfo = None

		if (os.environ.get('DISPLAY') or sys.platform[:3] == 'win' or
													sys.platform == 'darwin'):
			inout = file(os.devnull, 'r+')
		else:
			# for TTY programs, we need stdin/out
			inout = None

		# if possible, put the child precess in separate process group,
		# so keyboard interrupts don't affect child precess as well as
		# Python
		setsid = getattr(os, 'setsid', None)
		if not setsid:
			setsid = getattr(os, 'setpgrp', None)

		pipe = subprocess.Popen(cmdline, stdin=inout, stdout=inout,
								stderr=inout, close_fds=closefds,
								preexec_fn=setsid, startupinfo=startupinfo)

		# It is assumed that this kind of tools (gnome-open, kfmclient,
		# exo-open, xdg-open and open for OSX) immediately exit after lauching
		# the specific application
		returncode = pipe.wait()
		if hasattr(self, 'fixreturncode'):
			returncode = self.fixreturncode(returncode)
		return not returncode

	def open(self, filename):
		if isinstance(filename, basestring):
			cmdline = self.args + [filename]
		else:
			# assume it is a sequence
			cmdline = self.args + filename
		try:
			return self._invoke(cmdline)
		except OSError:
			return False

#Sysunsinn = True
# Platform support for Windows
if sys.platform[:3] == 'win':

	class Start(BaseController):
		'''Controller for the win32 start progam through os.startfile.'''
		
		def __init__(self, *args):
			#super(Controller, self).__init__(os.path.basename(args[0]))
			self.args = list(args)		

		def open(self, filename):
			try:
				os.startfile(filename)
			except WindowsError:
				# [Error 22] No application is associated with the specified
				# file for this operation: '<URL>'
				return False
			else:
				return True

	_controllers['windows-default'] = Start()
	_open = _controllers['windows-default'].open


# Platform support for MacOS
elif sys.platform == 'darwin':
	_controllers['open']= Controller('open')
	_open = _controllers['open'].open


# Platform support for Unix

else:
#elif Sysunsinn == False:

	import commands

	# @WARNING: use the private API of the webbrowser module
	from webbrowser import _iscommand

	class KfmClient(Controller):
		'''Controller for the KDE kfmclient program.'''

		def __init__(self, kfmclient='kfmclient'):
			super(KfmClient, self).__init__(kfmclient, 'exec')
			self.kde_version = self.detect_kde_version()

		def detect_kde_version(self):
			kde_version = None
			try:
				info = commands.getoutput('kde-config --version')

				for line in info.splitlines():
					if line.startswith('KDE'):
						kde_version = line.split(':')[-1].strip()
						break
			except (OSError, RuntimeError):
				pass

			return kde_version

		def fixreturncode(self, returncode):
			if returncode is not None and self.kde_version > '3.5.4':
				return returncode
			else:
				return os.EX_OK

	def detect_desktop_environment():
		'''Checks for known desktop environments

		Return the desktop environments name, lowercase (kde, gnome, xfce)
		or "generic"

		'''

		desktop_environment = 'generic'

		if os.environ.get('KDE_FULL_SESSION') == 'true':
			desktop_environment = 'kde'
		elif os.environ.get('GNOME_DESKTOP_SESSION_ID'):
			desktop_environment = 'gnome'
		else:
			try:
				info = commands.getoutput('xprop -root _DT_SAVE_MODE')
				if ' = "xfce4"' in info:
					desktop_environment = 'xfce'
			except (OSError, RuntimeError):
				pass

		return desktop_environment


	def register_X_controllers():
		if _iscommand('kfmclient'):
			_controllers['kde-open'] = KfmClient()

		for command in ('gnome-open', 'exo-open', 'xdg-open'):
			if _iscommand(command):
				_controllers[command] = Controller(command)

	def get():
		controllers_map = { \
			'gnome': 'gnome-open',
			'kde': 'kde-open',
			'xfce': 'exo-open',
		}

		desktop_environment = detect_desktop_environment()

		try:
			controller_name = controllers_map[desktop_environment]
			return _controllers[controller_name].open

		except KeyError:
			if _controllers.has_key('xdg-open'):
				return _controllers['xdg-open'].open
			else:
				return webbrowser.open


	if os.environ.get("DISPLAY"):
		register_X_controllers()
	_open = get()


def openfile(filename):
	'''Open a file or an URL in the registered default application.'''

	return _open(filename)


def _fix_addersses(**kwargs):
	for headername in ('address', 'to', 'cc', 'bcc'):
		try:
			headervalue = kwargs[headername]
			if not headervalue:
				del kwargs[headername]
				continue
			elif not isinstance(headervalue, basestring):
				# assume it is a sequence
				headervalue = ','.join(headervalue)

		except KeyError:
			pass
		except TypeError:
			raise TypeError('string or sequence expected for "%s", ' \
							'%s found' % (headername,
										  type(headervalue).__name__))
		else:
			translation_map = {'%': '%25', '&': '%26', '?': '%3F'}
			for char, replacement in translation_map.items():
				headervalue = headervalue.replace(char, replacement)
			kwargs[headername] = headervalue

	return kwargs


def mailto_format(**kwargs):
	# @TODO: implement utf8 option

	kwargs = _fix_addersses(**kwargs)
	parts = []
	for headername in ('to', 'cc', 'bcc', 'subject', 'body', 'attach'):
		if kwargs.has_key(headername):
			headervalue = kwargs[headername]
			if not headervalue:
				continue
			if headername in ('address', 'to', 'cc', 'bcc'):
				parts.append('%s=%s' % (headername, headervalue))
			else:
				headervalue = encode_rfc2231(headervalue) # @TODO: check
				parts.append('%s=%s' % (headername, headervalue))

	mailto_string = 'mailto:%s' % kwargs.get('address', '')
	if parts:
		mailto_string = '%s?%s' % (mailto_string, '&'.join(parts))

	return mailto_string


def mailto(address, to=None, cc=None, bcc=None, subject=None, body=None,
		   attach=None):
	'''Send an e-mail using the user's preferred composer.

	Open the user's preferred e-mail composer in order to send a mail to
	address(es) that must follow the syntax of RFC822. Multiple addresses
	may be provided (for address, cc and bcc parameters) as separate
	arguments.

	All parameters provided are used to prefill corresponding fields in
	the user's e-mail composer. The user will have the opportunity to
	change any of this information before actually sending the e-mail.

	address - specify the destination recipient
	cc	  - specify a recipient to be copied on the e-mail
	bcc	 - specify a recipient to be blindly copied on the e-mail
	subject - specify a subject for the e-mail
	body	- specify a body for the e-mail. Since the user will be able
			  to make changes before actually sending the e-mail, this
			  can be used to provide the user with a template for the
			  e-mail text may contain linebreaks
	attach  - specify an attachment for the e-mail. file must point to
			  an existing file

	'''

	mailto_string = mailto_format(**locals())
	return open(mailto_string)

